(c) C. Alex McLean 2007
Released under GPL version 2.  See the COPYING file or gnu.org for details.

> module Music where

> import Data.List (intersperse, nub, delete, elemIndex, transpose)

> import Util
> import Data.Maybe

> class (Music a) where
>     duration :: a -> Int
>     isSilent :: a -> Bool


> data Event a = Sound a
>              | Rest
>                deriving (Eq, Ord)

> instance Music (Event a) where
>     duration _ = 1
>     isSilent Rest = True
>     isSilent _ = False

> instance Show a => Show (Event a) where
>     show Rest = "_"
>     show (Sound x) = show x

> data Structure a = Atom (Event a)
>                  | Cycle [Structure a]
>                  | Polymetry [Structure a]
>                    deriving (Eq, Ord)

> type SStructure = Structure String

> instance Music (Structure a) where
>     duration (Atom _) = 1
>     duration (Cycle xs) = sum $ map duration xs
>     duration (Polymetry []) = 0
>     duration (Polymetry xs) = lcD $ map duration xs
>     isSilent (Atom x) = isSilent x
>     isSilent (Cycle xs) = and $ map isSilent xs
>     isSilent (Polymetry xs) = and $ map isSilent xs

> instance (Show a) => Show (Structure a) where
>     show (Atom x) = show x
>     show (Cycle xs) = concat $ intersperse " " $ map show xs
>     show (Polymetry xs)
>              = concat $ ("{" : (intersperse ", " $ map show xs)) ++ ["}"]

> silence = Cycle[Atom Rest]

> mapEvents :: ((Event a) -> (Event b)) -> ((Structure a) -> (Structure b))
> mapEvents f (Atom e) = Atom $ f e
> mapEvents f (Cycle xs) = Cycle $ map (mapEvents f) xs
> mapEvents f (Polymetry xs) = Polymetry $ map (mapEvents f) xs

> abstract :: (Structure String) -> (Structure Int)
> abstract s = mapEvents abstractEvent s
>     where lookup x = fromJust $ elemIndex x strings
>           strings = (nub $ vocableStrings s)
>           abstractEvent (Sound x) = Sound $ lookup x

> normalise :: (Structure String) -> (Structure String)
> normalise s@(Cycle []) = s
> normalise s = head $ normSort $ permute s

> structureToList :: (Structure a) -> [[Event a]]
> structureToList (Cycle xs) 
>     = concat $ map structureToList xs
> structureToList (Atom a) = [[a]]
> structureToList (Polymetry xs) 
>     = map concat $ transpose $ map structureToList xs

> normSort :: [(Structure String)] -> [(Structure String)]
> normSort [] = []
> normSort xs = map snd $ sortByFst $ zip (map abstract xs) xs

> permute :: Eq a => (Structure a) -> [(Structure a)]
> permute a@(Atom _) = [a]
> permute (Cycle xs) = map Cycle $ unravel $ map permute xs
> permute (Polymetry xs) = map Polymetry $ concat $ map perms possible
>     where possible = (unravel $ map permute xs)

Takes a list of lists containing possible items for each position, and
turns it into a list of possible lists

> unravel :: [[a]] -> [[a]]
> unravel [] = []
> unravel (a:[]) = map (\x -> [x]) a
> unravel (a:as) = concat $ map (\x -> (map (\y -> (y:x))) a) (unravel as)

> perms :: Eq a => [a] -> [[a]]
> perms [] = [[]]
> perms xs = [x:ys | x <- xs, ys <- perms (delete x xs)]

A Structure is unbalanced when it is or contains a Polymetry structure
containing Structures of different durations.  There are two ways to
balance such structures; expand or fill.  Both take the lowest common
multiple of the durations of the Structures within each unbalanced
Polymetry as the target duration.  With expand, each member is padded
out with Silence to arrive at the correct duration.  With fill, each
member is instead repeated.

> expand :: (Structure a) -> (Structure a)
> expand = balance expandTo

> fill :: (Structure a) -> (Structure a)
> fill = balance fillTo

> balance :: (Int -> (Structure a) -> (Structure a)) -> (Structure a) -> (Structure a)
> balance expander (Polymetry xs) = 
>     Polymetry $ map (expander newDuration) xs
>         where 
>           newDuration = foldl (lcm) 1 
>                         $ filter (\x -> x /= 0) 
>                         $ map duration xs
> balance _ m = m

> expandTo :: Int -> (Structure a) -> (Structure a)
> expandTo n s@(Cycle xs) = result
>     where result = Cycle $ map (\x -> expandTo (multiplier * (duration x)) x) xs
>           multiplier = n `div` (duration s)

> expandTo 1 s@(Atom _) = s
> expandTo n s@(Atom _) = Cycle (s : (replicate (n - 1) (Atom Rest)))
> expandTo n s@(Polymetry xs) = Polymetry $ map (expandTo n) xs

> fillTo :: Int -> (Structure a) -> (Structure a)
> fillTo n s@(Cycle xs) = 
>     Cycle $ concat $ (replicate multiplier xs)
>         where multiplier = n `div` (duration s)
> fillTo _ _ = undefined

> -- find lowest common duration (lcm but zeros are ignored)
> lcD :: [Int] -> Int
> lcD [] = 0
> lcD xs = lcD' $ filter (\x -> x /= 0) xs
>           where lcD' [] = 0
>                 lcD' ds = foldl (lcm) 1 ds

> isSound :: Event a -> Bool
> isSound (Sound _) = True
> isSound _ = False

> vocableStrings :: (Structure String) -> [String]
> vocableStrings x = map (\(Sound x) -> x) $ filter isSound $ events x

> events :: (Structure a) -> [Event a]
> events (Atom e) = [e]
> events (Cycle xs) = concat $ map events xs
> events (Polymetry xs) = concat $ map events xs

